#! /usr/bin/env python3
#
# Test(s) for ../template.py
#
# Notes:
# - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
# - * See test_python_ast.py for simple example of customization.
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by unittest_wrapper.py.
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/examples/tests/test_template.py
# - TODO: Remove TODO notes when stable.
#
# Warning:
# - The use of run_script as in test_01_data_file is an older style of testing.
#   It is better to directly invoke a helper class in the script that is independent
#   of the Script class based on Main. (See an example of this, see python_ast.py
#   and tests/tests_python_ast.py.)
#

## TODO1: [Warning] Make sure this template adhered to as much as possible. For,
## example, only delete todo comments not regular code, unless suggested in tip).
## In particular, it is critical that script_module gets initialized properly.

"""TODO: Tests for template module"""

# Standard modules
from typing import Optional

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
## TODO: from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:               module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
THE_MODULE = None
try:
    ## TODO: import mezcla.<module> as THE_MODULE
    ## -or-: import <module> as THE_MODULE
    pass                                ## TODO: delete
except:
    system.print_exception_info("<module> import") 
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        # Warning: see notes above about potential issues with run_script-based tests.
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        data = ["TEMP", "TODO", "DONE"]
        system.write_lines(self.temp_file, data)
        output = self.run_script(options="--TODO-arg", env_options="TODO_ENV=VAL",
                                 data_file=self.temp_file)
        self.do_assert(my_re.search(r"TODO Implement.me", output.strip()))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_helper(self):
        """Test for helper class"""
        debug.trace(4, f"TestIt.test_02_helper(); self={self}")
        TODO_var: Optional[bool] = None
        helper = THE_MODULE.Helper(TODO_var)
        self.do_assert(not helper.process("TODO: some arg"))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_03_captured_stdout(self):
        """Ensure that stdout and/or stderr captured properly"""
        debug.trace(4, "test_03_captured_IO()")
        helper = THE_MODULE.Helper()
        self.do_assert(not helper.process("TODO: some arg"))
        stdout = self.get_stdout()
        self.do_assert(my_re.search(r"Error.*Implement.me", stdout.strip()))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_04_captured_stderr(self):
        """Ensure that stderr captured properly"""
        debug.trace(4, "test_04_captured_stderr()")
        self.patch_trace_level(5)
        THE_MODULE.Helper()
        stderr = self.get_stderr()
        self.do_assert(my_re.search(r"Helper.*instance", stderr.strip()))
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
# TODO: # -*- coding: utf-8 -*-
## TODO: handle case when env installed elsewhere (e.g., maldito mac)
## #! env python
#
# TODO what the script does (detailed)
# -or-
# Based on following:
#   TODO: url
#

"""
TODO: what module does (brief)

Sample usage:
   echo $'TODO:task1\\nDONE:task2' | {script} --TODO-arg --
"""

# Standard modules
from typing import Optional

# Installed modules
## TODO: import numpy as np

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main, FILENAME
from mezcla.my_regex import my_re
from mezcla import system
## TODO:
## from mezcla import data_utils as du
## TODO2: streamline imports by exposing common functions, etc. in mezcla
##
## Optional:
## # Increase trace level for regex searching, etc. (e.g., from 6 to 7)
## my_re.TRACE_LEVEL = debug.QUITE_VERBOSE
debug.trace(5, f"global __doc__: {__doc__}")
debug.assertion(__doc__)

# Constants
TL = debug.TL
## TODO: Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
## Note: Run following in Emacs to interactively replace TODO_ARG with option label
##    M-: (query-replace-regexp "todo\\([-_]\\)arg" "arg\\1name")
## where M-: is the emacs keystroke short-cut for eval-expression.
##
## TODO: TODO_BOOL_OPT = "todo-bool-option"
## TODO: TODO_TEXT_OPT = "todo-text-option"

# Environment options
# Notes:
# - These are just intended for internal options, not for end users.
# - They also allow for enabling options in one place rather than four
#   when using main.Main (e.g., [Main member] initialization, run-time
#   value, and argument spec., along with string constant definition).
# WARNING: To minimize environment comflicts with other programs make the names
# longer such as two or more tokens (e.g., "FUBAR" => "FUBAR_LEVEL").
#
## TODO_FUBAR = system.getenv_bool(
##     "TODO_FUBAR", False,
##     description="TODO:Fouled Up Beyond All Recognition processing")

#-------------------------------------------------------------------------------

class Helper:
    """TODO: class for doing ..."""

    def __init__(self, _arg=None, **kwargs) -> None:
        """Initializer: TODO_arg desc"""
        debug.trace_expr(TL.VERBOSE, _arg, kwargs, prefix="in Helper.__init__: ")
        self._arg = _arg                # TODO: revise
        self.TODO: Optional[bool] = None
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def process(self, _arg) -> bool:
        """TODO: Process _ARG to do ..."""
        ## NOTE: print used for sake of unit test (see examples/tests/test_template.py)
        print("Error: TODO Implement me!")
        return False

#-------------------------------------------------------------------------------

def main() -> None:
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Parse command line options, show usage if --help given
    # TODO: manual_input=True; short_options=True
    # Note: Uses Main without subclassing, so some methods are stubs (e.g., run_main_step).
    main_app = Main(
        description=__doc__.format(script=gh.basename(__file__)),
        ## TODO: boolean_options=[(TODO_BOOL_OPT, "TODO desc1")],
        ## TODO: text_options=[(TODO_TEXT_OPT, "TODO desc2")],
        ## TODO: positional_arguments=[FILENAME, ALT_FILENAME], 
    )
    debug.assertion(main_app.parsed_args)
    ## TODO_opt1 = main_app.get_parsed_option(TODO_BOOL_OPT)

    helper = Helper()
    helper.process("TODO: some argument")
    
    ## TODO:
    ## ALT TODO:
    ## for line in main_app.read_entire_input().splitlines():
    ##     helper.process(TODO_fn(line))
    ## -or-
    ## helper.process( main_app.read_entire_input())

    ## Make sure no TODO_vars above (i.e., in namespace); TODO: delete check when stable
    debug.assertion(not any(my_re.search(r"^TODO_", m, my_re.IGNORECASE)
                            for m in dir(main_app)))
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    debug.trace(5, f"module __doc__: {__doc__}")
    debug.assertion("TODO:" not in __doc__)
    main()



#! /usr/bin/env python3
#
# Cartoonifies
# - Based on following:
#   https://pypi.org/project/cartoon-diffusion
# - Code prepared with GPT 4.1 (with a bit of trial and error).
#

"""
Convert images to cartoon renditions using cartoon-diffusion

Sample usage:
   {script} --save ~/Downloads/amlo.jpeg

   {script} dylan.png > cartoonified-dylan.png
"""

# Standard modules
from pathlib import Path

# Installed modules
## NOTE: import delayed so --help usage shown sooner (and sans tensorflow junk!)
## OLD: from cartoon_diffusion import CartoonifyDiffusionPipeline

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system

# Constants
TL = debug.TL
SAVE_OPT = "save"
PATH_OPT = "image-path"
OUTPUT_PREFIX = "cartoonified-"

#-------------------------------------------------------------------------------

class CartoonifyHelper:
    """Helper class for converting images to cartoon style"""

    def __init__(self):
        """Initializer: loads the cartoon diffusion pipeline"""
        debug.trace(TL.VERBOSE, f"CartoonifyHelper.__init__(): self={self}")
        # pylint: disable=import-outside-toplevel
        from cartoon_diffusion import CartoonifyDiffusionPipeline
        self.pipeline = CartoonifyDiffusionPipeline.from_pretrained("wizcodes12/image_to_cartoonify")
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def process(self, image_path):
        """Convert image at IMAGE_PATH to cartoon style"""
        cartoon = self.pipeline(image_path)
        return cartoon

    def save_with_prefix(self, cartoon, original_path):
        """Save CARTOON image with prefix based on ORIGINAL_PATH"""
        path = Path(original_path)
        output_path = path.parent / f"{OUTPUT_PREFIX}{path.name}"
        cartoon.save(str(output_path))
        return str(output_path)

#-------------------------------------------------------------------------------

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Parse command line options
    main_app = Main(
        description=__doc__.format(script=gh.basename(__file__)),
        positional_arguments=[(PATH_OPT, "Path to input image")],
        boolean_options=[(SAVE_OPT, "Save output to file with 'cartoonified-' prefix")],
        manual_input=True
    )
    debug.assertion(main_app.parsed_args)
    
    # Get arguments
    save_to_file = main_app.get_parsed_option(SAVE_OPT)
    image_path = main_app.get_parsed_argument(PATH_OPT)
    
    # Process image
    helper = CartoonifyHelper()
    cartoon = helper.process(image_path)
    
    # Output result
    if save_to_file:
        output_path = helper.save_with_prefix(cartoon, image_path)
        print(f"Created {output_path}")
    else:
        # For stdout output, just save to a temp file and let shell handle piping
        cartoon.save("/dev/stdout")
    
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    main()



#! /usr/bin/env python3
#
# Test(s) for ../<module>.py
#
# Notes:
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/tests/test_<module>.py
#

"""Tests for introspection module"""

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
THE_MODULE = None
try:
    import mezcla.introspection as THE_MODULE
except:
    system.print_exception_info("introspection import") 

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_simple_introspection(self):
        """Test for simple introspection"""
        debug.trace(4, f"TestIt.test_01_simple_introspection(); self={self}")
        fubar = 123.321
        fubar_expr = THE_MODULE.intro.format(fubar)
        self.do_assert(my_re.search("fubar.*123.321", fubar_expr))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_multiline_introspection(self):
        """Test for multi-line introspection"""
        debug.trace(4, f"TestIt.test_02_multiline_introspection(); self={self}")
        multiline_value_expr = THE_MODULE.intro.format(
            2
            +
            2
            ==
            5)
        debug.trace_expr(5, multiline_value_expr)
        self.do_assert(my_re.search(r"2.*\+.*2.*==.*5", multiline_value_expr,
                                    flags=my_re.DOTALL|my_re.MULTILINE))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_03_prefix(self):
        """Test for prefix added to introspection result"""
        debug.trace(4, f"TestIt.test_03_prefix(); self={self}")
        var = 456
        var_expr = THE_MODULE.intro.format(var, _prefix="here: ")
        assert my_re.search("here: var.*456", var_expr)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_04_max_len(self):
        """Test for introspection truncation"""
        debug.trace(4, f"TestIt.test_04_max_len(); self={self}")
        var = "-" * 123
        var_expr = THE_MODULE.intro.format(var, max_len=4)
        ## TODO2: assert my_re.search("var='----\.\.\.'", var_expr)
        ##                                            ^ (i.e., make sure quote closed)
        assert my_re.search(r"var='----\.\.\.", var_expr)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_05_quirks(self):
        """Test for known introspection quirks"""
        debug.trace(4, f"TestIt.test_05_quirks(); self={self}")
        # pylint: disable=unnecessary-lambda-assignment
        indirect_introspection = lambda x: THE_MODULE.intro.format(x, indirect=True)

        # Quirk 1: Unable to resolve variable defined on same line as call
        #
        ## NOTE: attempts to reproduce following:
        ##   In [150]: fu=123; MezclaDebugger().format(fu)
        ##   Warning: unable to resolve call node: args=(123,)
        ##   Out[150]: '123\n'
        ## versus:
        ##   In [151]: MezclaDebugger().format(fu)
        ##   Out[151]: 'fu=123\n'
        #
        ## TODO:
        ## pylint: disable=multiple-statements
        ## same_line_var = 123; expr = indirect_introspection(same_line_var)
        ## assert my_re.search(rf"^{same_line_var}", expr)
        #
        other_line_var = 321
        expr = indirect_introspection(other_line_var)
        assert my_re.search(rf"^other_line_var={other_line_var}", expr)

        # Quirk 2: Unable to resolve expressions embedded in function calls
        num = 7
        add1 = lambda num: 1 + num      
        assert not my_re.search(r"^add.*add.*num.*=", indirect_introspection(add1(add1(num))))
        #
        add1_add1_num = add1(add1(num))
        expr = indirect_introspection(add1_add1_num)
        assert my_re.search(rf"^add1_add1_num={add1_add1_num}", expr)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../check_html_javascript.py
#
# Notes:
# - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
# - * See test_python_ast.py for simple example of customization.
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by unittest_wrapper.py.
# - This can be run as follows (e.g., from root of repo):
#   $ pytest ./mezcla/tests/test_check_html_javascript.py
#
# Warning:
# - The use of run_script as in test_01_data_file is an older style of testing.
#   It is better to directly invoke a helper class in the script that is independent
#   of the Script class based on Main. (See an example of this, see python_ast.py
#   and tests/tests_python_ast.py.)
# - Moreover, debugging tests with run_script is complicated because a separate
#   process is involved (e.g., with separate environment variables.)
# - See discussion of SUB_DEBUG_LEVEL in unittest_wrapper.py for more info.
# - TODO: Feel free to delete this warning as well as the related one below.
#


"""Tests for check_html_javascript module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Two references are used for the module to be tested:
#    THE_MODULE:               module object (e.g., <module 'mezcla.main' ...>)
#    TestIt.script_module:     dotted module path (e.g., "mezcla.main")
THE_MODULE = None
try:
    import mezcla.check_html_javascript as THE_MODULE
except:
    system.print_exception_info("check_html_javascript import") 
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(r"\btemplate.py$", __file__):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

# Constants
    
SIMPLE_HTML_TEXT = """
    <!DOCTYPE html>
    <html>
      <head>
        <title>Dummy HTML file</title>
        <script>
           console.log("in script");
           text = "fubar";
           text -= 666
	   console.log("out script");
        </script>
      </head>
    
      <body>
         Dummy
      </body>
    </html>    
"""

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")

        # Run script over simple HTML data and get outut
        system.write_file(self.temp_file, SIMPLE_HTML_TEXT)
        output = self.run_script(options="", data_file=self.temp_file)
        output = output.strip()
        
        # Warning from jslint
        # 
        self.do_assert(my_re.search(r"Undeclared.*text", output))
        
        # Warning from jshint
        self.do_assert(my_re.search(r"Missing semicolon", output))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_02_strict_mode(self):
        """Make sure strict mode used"""
        debug.trace(4, f"TestIt.ttest_02_strict_mode(); self={self}")
        self.do_assert("strict" in THE_MODULE.SAFEMODE_HEADER)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# POE client written with the help of POE - Updated for OpenAI-compatible API
#
# Example usage:
#   client = POEClient(base_url="https://api.poe.com/v1", api_key="your_api_key")
#   response = client.ask(model="gpt-3.5-turbo", question="What is the capital of France?")
#   print(response)
#

"""
A Python client for interacting with POE (Platform for Open Exploration) LLMs via OpenAI-compatible API.
This module provides a class for basic Q&A functionality and extensibility for advanced features.

Sample usage:
    POE_API="..." {script} --command "What is the color of MAGA?"
"""

# Standard modules
from typing import Any, Dict, Optional, List

# Installed modules
import requests

# Local modules
# TODO: def mezcla_import(name): ... components = eval(name).split(); ... import nameN-1.nameN as nameN
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system


# Environment options
POE_API = system.getenv_value(
    "POE_API", None,
    desc="API key for POE")
POE_MODEL = system.getenv_value(
    ## OLD: "POE_MODEL", "GPT-4.1-nano",
    "POE_MODEL", None,
    desc="Default model for POE")
POE_URL = system.getenv_text(
    "POE_URL", "https://api.poe.com/v1",
    desc="Base URL for POE API")
POE_TIMEOUT = system.getenv_float(
    "POE_TIMEOUT", 30,
    desc="Timeout for POE API call")


class POEClient:
    """
    A Python client for interacting with POE LLMs via OpenAI-compatible API.
    Supports basic Q&A functionality and is designed to be extensible.
    """

    def __init__(self, base_url: Optional[str] = None, api_key: Optional[str] = None, timeout: Optional[float] = None, model: Optional[str] = None):
        """
        Initialize the client with the base URL of the POE API and an API key.

        Args:
            base_url (str): The base URL of the POE API.
            api_key (str): The API key for authentication.
            timeout (float): Timeout for HTTP requests in seconds (e.g., 30).
            model (str): Default model to use.
        """
        debug.trace_expr(6, base_url, api_key, timeout, model,
                         prefix="in __init__: ")
        if base_url is None:
            base_url = POE_URL
        self.base_url = base_url.rstrip("/")
        if api_key is None:
            api_key = POE_API
        self.api_key = api_key
        if timeout is None:
            timeout = POE_TIMEOUT
        self.timeout = timeout
        if model is None:
            debug.assertion(POE_MODEL)
            model = POE_MODEL
        self.model = model
        self.headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def _send_request(
        self, endpoint: str, payload: Optional[Dict[str, Any]] = None, method: str = "POST"
    ) -> Dict[str, Any]:
        """
        Internal method to send a request to the POE API.

        Args:
            endpoint (str): The API endpoint (relative to the base URL).
            payload (Optional[Dict[str, Any]]): The request payload (optional).
            method (str): The HTTP method ("POST" or "GET").

        Returns:
            Dict[str, Any]: The API response as a dictionary.

        Raises:
            RuntimeError: If the request fails due to an HTTP or connection error.
        """
        debug.trace_expr(5, endpoint, payload, method,
                         prefix="in _send_request: ")
        
        url = f"{self.base_url}/{endpoint}"
        
        try:
            if method.upper() == "POST":
                response = requests.post(
                    url, headers=self.headers, json=payload, timeout=self.timeout
                )
            elif method.upper() == "GET":
                response = requests.get(
                    url, headers=self.headers, params=payload, timeout=self.timeout
                )
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")
            
            debug.trace_object(6, response)
            response.raise_for_status()  # Raise HTTPError for bad responses (4xx and 5xx)
            
            result = response.json()
            debug.trace(5, f"_send_request() => {result}")
            return result
            
        except requests.exceptions.HTTPError as http_err:
            debug.raise_exception(6)
            error_details = ""
            try:
                error_response = http_err.response.json()
                error_details = f" - {error_response}"
            except:
                pass
            raise RuntimeError(
                f"HTTP error occurred: {http_err.response.status_code} {http_err.response.reason}{error_details}"
            ) from http_err
        except requests.exceptions.RequestException as req_err:
            debug.raise_exception(6)
            raise RuntimeError(f"Error in API request: {req_err}") from req_err

    def ask(self, question: str, model: Optional[str] = None, context: Optional[str] = None, 
            temperature: float = 0.7, max_tokens: Optional[int] = None) -> str:
        """
        Send a question to the specified model with optional context using OpenAI chat completions format.

        Args:
            question (str): The question to ask the model.
            model (Optional[str]): The name of the model to use (uses default if None).
            context (Optional[str]): Additional context for the model (optional).
            temperature (float): Controls randomness in the response (0.0 to 2.0).
            max_tokens (Optional[int]): Maximum number of tokens in the response.

        Returns:
            str: The model's response.
        """
        debug.trace_expr(6, model, question, context, temperature, max_tokens,
                         prefix="in ask: ")
        
        if model is None:
            model = self.model
            
        messages = []
        if context:
            messages.append({"role": "system", "content": context})
        messages.append({"role": "user", "content": question})
        
        payload = {
            "model": model,
            "messages": messages,
            "temperature": temperature,
        }
        
        if max_tokens is not None:
            payload["max_tokens"] = max_tokens

        response = self._send_request("chat/completions", payload)
        
        # Extract the response content from OpenAI format
        try:
            result = response["choices"][0]["message"]["content"]
        except (KeyError, IndexError) as e:
            debug.trace(4, f"Unexpected response format:\n\t{response}\n\t{e}")
            result = response.get("output", str(response))
        debug.trace(5, f"ask() => {result}")
        return result

    def create_chat_completion(self, messages: List[Dict[str, str]], model: Optional[str] = None,
                              temperature: float = 0.7, max_tokens: Optional[int] = None,
                              stream: bool = False, **kwargs) -> Dict[str, Any]:
        """
        Create a chat completion using the OpenAI chat completions format.

        Args:
            messages (List[Dict[str, str]]): List of messages in OpenAI format.
            model (Optional[str]): The name of the model to use.
            temperature (float): Controls randomness in the response.
            max_tokens (Optional[int]): Maximum number of tokens in the response.
            stream (bool): Whether to stream the response (not yet implemented).
            **kwargs: Additional parameters to pass to the API.

        Returns:
            Dict[str, Any]: The complete API response.
        """
        debug.trace_expr(6, messages, model, temperature, max_tokens, stream,
                         prefix="in create_chat_completion: ")
        
        if model is None:
            model = self.model
            
        payload = {
            "model": model,
            "messages": messages,
            "temperature": temperature,
            **kwargs
        }
        
        if max_tokens is not None:
            payload["max_tokens"] = max_tokens
            
        if stream:
            payload["stream"] = stream
            # TODO: Implement streaming support
            debug.trace(3, "Warning: Streaming not yet implemented")

        result = self._send_request("chat/completions", payload)
        debug.trace(5, f"create_chat_completion() => {result}")
        return result

    def list_models(self) -> Dict[str, Any]:
        """
        List available models.

        Returns:
            Dict[str, Any]: The list of available models.
        """
        debug.trace(5, "in list_models")
        return self._send_request("models", method="GET")

    def call_function(self, function_name: str, arguments: Dict[str, Any], 
                     model: Optional[str] = None, context: Optional[str] = None) -> Any:
        """
        Call a specific function using function calling in chat completions.

        Args:
            function_name (str): The name of the function to call.
            arguments (Dict[str, Any]): The arguments for the function.
            model (Optional[str]): The name of the model to use.
            context (Optional[str]): Additional context for the function call.

        Returns:
            Any: The function's output.
        """
        debug.trace_expr(6, model, function_name, arguments, context,
                         prefix="in call_function: ")
        
        if model is None:
            model = self.model

        messages = []
        if context:
            messages.append({"role": "system", "content": context})
        messages.append({
            "role": "user", 
            "content": f"Please call the function {function_name} with these arguments: {arguments}"
        })

        # Define the function for function calling
        functions = [{
            "name": function_name,
            "description": f"Call the {function_name} function",
            "parameters": {
                "type": "object",
                "properties": {key: {"type": "string"} for key in arguments.keys()},
                "required": list(arguments.keys())
            }
        }]

        payload = {
            "model": model,
            "messages": messages,
            "functions": functions,
            "function_call": {"name": function_name}
        }

        response = self._send_request("chat/completions", payload)
        
        try:
            function_call = response["choices"][0]["message"].get("function_call")
            if function_call:
                response = function_call.get("arguments", {})
            else:
                response = response["choices"][0]["message"]["content"]
        except (KeyError, IndexError):
            response = response.get("output")
        
        debug.trace(5, f"call_function() => {response}")
        return response

    def extend(self, extension_name: str, params: Dict[str, Any], model: Optional[str] = None) -> Dict[str, Any]:
        """
        Extend the model's capabilities using chat completions with custom instructions.

        Args:
            extension_name (str): The name of the extension to apply.
            params (Dict[str, Any]): Parameters for the extension.
            model (Optional[str]): The name of the model to use.

        Returns:
            Dict[str, Any]: The extension's result.
        """
        debug.trace_expr(5, model, extension_name, params,
                         prefix="in extend: ")
        
        if model is None:
            model = self.model

        # Convert extension call to a chat completion
        system_message = f"You are operating in {extension_name} mode with the following parameters: {params}"
        user_message = f"Please process this request using the {extension_name} extension."

        messages = [
            {"role": "system", "content": system_message},
            {"role": "user", "content": user_message}
        ]

        response = self.create_chat_completion(messages, model=model)
        debug.trace(5, f"extend() => {response}")
        return response


def main():
    """Entry point for testing"""
    debug.trace(4, "POE Client - OpenAI Compatible API")
    
    # Parse command line options, show usage if --help given
    LIST_MODELS_ARG = "list-models"
    COMMAND_ARG = "command"
    STDIO_ARG = "stdio"
    main_app = Main(
        description=__doc__.format(script=gh.basename(__file__)),
        boolean_options=[(LIST_MODELS_ARG, "List available LLM models"),
                         (STDIO_ARG, "Use stdin for command (and plain stdout for output)")],
        text_options=[(COMMAND_ARG, "Command or question for LLM")],
    )
    debug.assertion(main_app.parsed_args)
    list_models = main_app.get_parsed_option(LIST_MODELS_ARG)
    llm_command = main_app.get_parsed_option(COMMAND_ARG)
    use_stdio = main_app.get_parsed_option(STDIO_ARG)
    if use_stdio:
        debug.assertion(not llm_command)
        llm_command = main_app.read_entire_input()
    
    # Example usage
    if not POE_API:
        system.exit("Error: POE_API environment variable not set. Cannot run client.")
    try:
        client = POEClient()
        
        # Test model listing
        if list_models:
            models = client.list_models()
            print(f"Available models:\n\t{models}")
            
        # Test basic ask functionality
        if llm_command:
            response = client.ask(llm_command)
            print(f"Response:\n\t{response}" if not use_stdio else response)
        
    except:
        system.print_exception_info("Error testing client")

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(6)
    debug.trace(5, f"module __doc__: {__doc__}")
    main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Test(s) for tpo_common.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_tpo_common.py
#
# TODO:
# - Address commonly used debugging functions (e.g., debug_print) by redirecting output (via remapping sys.stderr to a file) and then checking file contents.
# - add tests for normalize_unicode, ensure_unicode and other problematic functions
# TODO2:
# - Remove most of the trace.Trace usages as that usually is overkill:
#   capturing stderr is sufficient!
#
# TODO3:
# - Drop tests that are similar to those in test_debug.py (given that
#   tpo_common.py mosly implemented in terms of debug.py).
#
# Important:
# - Most of the methods on tpo_common was moved to newer modules, such
#   as system.py, debug.py etc.
# - Using the newer module tests to these moved methods could be useful to avoid
#   repeated tests but this could create conflicts with modified methods.
#

"""Tests for tpo_common module"""

# Standard modules
import sys
import re
## OLD: from datetime import datetime
import os
import math
import trace
import pickle

# Installed modules
import pytest

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla.unittest_wrapper import trap_exception

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	                global module object
#    TestIt.script_module:              path to file
import mezcla.tpo_common as THE_MODULE

FUBAR = 101	# sample global for test_format
FOOBAR = 12     # likewise
JOSE = "José"   # UTF-8 encoded string
UTF8_BOM = "\xEF\xBB\xBF"

class TestTpoCommon(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def test_set_debug_level(self):
        """Ensure set_debug_level works as expected"""
        debug.trace(4, "test_set_debug_level()")
        self.patch_trace_level(999)
        assert (THE_MODULE.debugging_level() == 999) or (not __debug__)

    def test_debugging_level(self):
        """Ensure debugging_level works as expected"""
        debug.trace(4, "test_debugging_level()")
        self.patch_trace_level(999)
        assert (THE_MODULE.debugging_level() == 999) or (not __debug__)

    def test_debug_trace_without_newline(self):
        """Ensure debug_trace_without_newline works as expected"""
        debug.trace(4, "test_debug_trace_without_newline()")
        text = 'test debug trace without newline'

        # test underlying function is being called
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.debug_trace_without_newline, (text), level=1)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()
        assert re.search(r'modulename: debug, funcname: trace', out)

        # test behaviour of functions
        assert text in error
        assert f'{text}\n' not in error

    @pytest.mark.xfail
    def test_debug_trace_basic(self):
        """Test basic debug_trace_basic usage"""
        debug.trace(4, "test_debug_trace()")
        text = 'test debug trace without newline'

        # test underlying function is being called
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.debug_trace, (text), level=1)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()
        assert re.search(r'modulename: debug, funcname: trace', out)

        # test behaviour of functions
        assert f'{text}\n' in error

    @pytest.mark.xfail
    def test_debug_print(self):
        """Ensure debug_print works as expected"""
        debug.trace(4, "test_debug_print()")
        self.patch_trace_level(2)
        text = 'test debug trace without newline'

        # test underlying function is being called
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.debug_print, (text),
                       level=1, skip_newline=False)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()
        assert re.search(r'modulename: debug, funcname: trace', out)

        # test behaviour of functions
        assert f'{text}\n' in error

        # test with skip_newline
        self.clear_stdout_stderr()
        tracer.runfunc(THE_MODULE.debug_print, (text),
                       level=1, skip_newline=True)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()
        assert re.search(r'modulename: debug, funcname: trace', out)

        # test behaviour of functions
        assert text in error
        assert f'{text}\n' not in error

    @pytest.mark.xfail
    def test_debug_format(self):
        """Ensure debug_format works as expected"""
        debug.trace(4, "test_debug_format()")
        text = "test {function}"
        THE_MODULE.debug_format(text=text, level=4, skip_newline=False,
                                function="debug_format")
        stderr = self.get_stderr()
        assert "test debug_format" in stderr

    def test_debug_timestamp(self):
        """Ensure debug_timestamp works as expected"""
        debug.trace(4, "test_debug_timestamp()")
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.debug_timestamp)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        captured = self.get_stdout()
        assert re.search(r'modulename: debug, funcname: timestamp', captured)

        # test behaviour of function
        #
        ## BAD (the following timestamps might differ in seconds due to millisecond wrapping):
        ## actual, expected = THE_MODULE.debug_timestamp(), str(datetime.now())
        ## # truncating the timestamp to eliminate milisecond difference in calculation
        ## assert actual[:22] == expected[:22]
        #
        # EX: "2025-08-05 20:49:22.053904" => "2025-08-05 20:49:22.053"
        #     "01234567890123456789012"
        actual = THE_MODULE.debug_timestamp()
        assert my_re.sub("[0-9]", "N", actual[:23]) == "NNNN-NN-NN NN:NN:NN.NNN"

    def test_debug_raise(self):
        """Ensure debug_raise works as expected"""
        debug.trace(4, "test_debug_raise()")

        tracer = trace.Trace(countfuncs=1)
        with pytest.raises(RuntimeError):
            tracer.runfunc(THE_MODULE.debug_raise)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        captured = self.get_stdout()
        assert re.search(r'modulename: debug, funcname: raise_exception', captured)

    def test_trace_array(self):
        """Ensure trace_array works as expected"""
        debug.trace(4, "test_trace_array()")
        array = ['test', 'trace', 'array']
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.trace_array, (array), level=1)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()
        assert re.search(r'modulename: debug, funcname: trace_values', out)

        for i, item in enumerate(array):
            assert f"{i}: {item}" in error

    def test_trace_object(self):
        """Ensure trace_object works as expected"""
        debug.trace(4, "test_trace_object()")

        class TestObj:
            """Class for trace_object"""
            __var__ = 1

            def test_method(self):
                """Dummy method"""
                pass

        obj = TestObj()
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.trace_object, (obj),
                       level=1, show_methods_etc=True, show_private=True)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()
        assert re.search(r'modulename: debug, funcname: trace_object', out)

        assert '__var__: 1' in error
        assert 'test_method' in error

    @pytest.mark.xfail
    def test_trace_value(self):
        """Ensure trace_value works as expected"""
        debug.trace(4, "test_trace_value()")
        value = "test value"
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.trace_value, (value), level=1)

        # redirect write_results tot temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()
        assert re.search(r'modulename: tpo_common, funcname: debug_print', out)
        assert value in error

        tracer.runfunc(THE_MODULE.trace_value, (value), label='TPO', level=1)

        # redirect write_results to temp file
        tracer.results().write_results(coverdir=temp)

        out, error = self.get_stdout_stderr()

        assert f"TPO: {value}" in error

    @pytest.mark.xfail
    def test_trace_current_context(self):
        """Ensure trace_current_context works as expected"""
        debug.trace(4, "test_trace_current_context()")
        # note: level should be higher than level below
        base_trace_level = 4
        self.patch_trace_level(base_trace_level + 1)

        test_var = 1                    # pylint: disable=unused-variable
        # check that local and global variables are traced
        THE_MODULE.trace_current_context(level=base_trace_level, label='TPO',
                                         show_methods_etc=False, max_value_len=9999)
        error = self.get_stderr()
        debug.trace_expr(4, error)
        assert 'TPO' in error
        assert 'test_var' in error
        assert 'FOOBAR' in error

        # assert debug function is being called
        tracer = trace.Trace(countfuncs=1)
        tracer.runfunc(THE_MODULE.trace_current_context, level=2, label='TPO', show_methods_etc=False)

        # redirect write_results to temp file
        temp = self.get_temp_file()
        tracer.results().write_results(coverdir=temp)

        stdout = self.get_stdout()
        assert re.search(r'modulename: debug, funcname: trace_current_context', stdout)

    @pytest.mark.xfail
    def test_during_debugging(self):
        """Ensure during_debugging works as expected"""
        debug.trace(4, "test_during_debugging()")
        assert THE_MODULE.during_debugging()
        assert not THE_MODULE.during_debugging(False)

    def test_debugging(self):
        """Ensure debugging works as expected"""
        debug.trace(4, "test_debugging()")
        self.patch_trace_level(4)
        assert THE_MODULE.debugging(2)
        assert THE_MODULE.debugging(4)
        assert not THE_MODULE.debugging(6)

    def test_detailed_debugging(self):
        """Ensure detailed_debugging works as expected"""
        debug.trace(4, "test_detailed_debugging()")
        self.patch_trace_level(2)
        assert not THE_MODULE.detailed_debugging()
        self.patch_trace_level(4)
        assert THE_MODULE.detailed_debugging()
        self.patch_trace_level(6)
        assert THE_MODULE.detailed_debugging()

    def test_verbose_debugging(self):
        """Ensure verbose_debugging works as expected"""
        debug.trace(4, "test_verbose_debugging()")
        self.patch_trace_level(2)
        assert not THE_MODULE.verbose_debugging()
        self.patch_trace_level(5)
        assert THE_MODULE.verbose_debugging()
        self.patch_trace_level(7)
        assert THE_MODULE.verbose_debugging()

    def test_to_string(self):
        """Ensure to_string works as expected"""
        debug.trace(4, "test_to_string()")
        assert THE_MODULE.to_string(123) == "123"
        assert THE_MODULE.to_string("\u1234") == "\u1234"
        assert THE_MODULE.to_string(None) == "None"

    def test_normalize_unicode(self):
        """Ensure normalize_unicode works as expected"""
        debug.trace(4, "test_normalize_unicode()")
        assert THE_MODULE.normalize_unicode("ASCII") == "ASCII"
        assert THE_MODULE.normalize_unicode(UTF8_BOM) == UTF8_BOM
        ## TODO: assert "Jos\xc3\xa9", THE_MODULE.normalize_unicode(JOSE)
        ## TODO: add tests for sys.version_info.major < 3
        ## assert THE_MODULE.normalize_unicode('\u1234') == '\xe1\x88\xb4'

    def test_ensure_unicode(self):
        """Ensure ensure_unicode works as expected"""
        debug.trace(4, "test_ensure_unicode()")
        assert THE_MODULE.ensure_unicode("ASCII") == "ASCII"
        assert "Jos\xe9" == THE_MODULE.ensure_unicode(JOSE)
        ## TODO: assert THE_MODULE.ensure_unicode(UTF8_BOM) == '\ufeff'
        ## TODO: add tests for sys.version_info.major < 3
        ## assert THE_MODULE.ensure_unicode('\xe1\x88\xb4') == '\u1234'

    def test_print_stderr(self):
        """Ensure print_stderr works as expected"""
        debug.trace(4, "test_print_stderr()")
        # ensure stderr is not being redirected
        self.monkeypatch.setattr(THE_MODULE, 'stderr', sys.stderr)

        message = "this is error"
        THE_MODULE.print_stderr(message)
        error = self.get_stderr()
        assert message in error

    def test_redirect_stderr(self):
        """Ensure redirect_stderr works as expected"""
        debug.trace(4, "test_redirect_stderr()")
        # ensure stderr is not being redirected
        self.monkeypatch.setattr(THE_MODULE, 'stderr', sys.stderr)

        temp = self.get_temp_file()
        THE_MODULE.redirect_stderr(temp)
        THE_MODULE.print_stderr("stderr redirected")
        THE_MODULE.restore_stderr()
        assert "stderr redirected" in gh.read_file(temp)

    @pytest.mark.xfail
    def test_restore_stderr(self):
        """Ensure restore_stderr works as expected"""
        debug.trace(4, "test_restore_stderr()")
        # ensure stderr is already being redirected
        self.monkeypatch.setattr(THE_MODULE, 'stderr',
                                 system.open_file(self.get_temp_file()))
        THE_MODULE.restore_stderr()
        assert sys.stderr == THE_MODULE.stderr

    def test_setenv(self):
        """Ensure setenv works as expected"""
        debug.trace(4, "test_setenv()")
        THE_MODULE.setenv('NEW_TEST_ENV_VAR', 'the gravity is 10, pi is 3')
        assert THE_MODULE.getenv('NEW_TEST_ENV_VAR') == 'the gravity is 10, pi is 3'

    def test_chomp(self):
        """Ensure chomp works as expected"""
        debug.trace(4, "test_chomp()")
        assert THE_MODULE.chomp("abc" + os.linesep) == "abc"
        assert THE_MODULE.chomp("http://localhost/", "/") == "http://localhost"

    @trap_exception
    def test_register_env_option(self):
        """Ensure register_env_option works as expected"""
        debug.trace(4, "test_register_env_option()")

        THE_MODULE.env_options = {}
        THE_MODULE.env_defaults = {}

        THE_MODULE.register_env_option(
            var='VAR_STRING',
            description='this is a string variable',
            default='empty'
        )

        self.do_assert(THE_MODULE.env_options['VAR_STRING'], 'this is a string variable')
        self.do_assert(THE_MODULE.env_defaults['VAR_STRING'], 'empty')
        self.do_assert(len(THE_MODULE.env_options) == 1)
        self.do_assert(len(THE_MODULE.env_defaults) == 1)

    def set_test_env_var(self):
        """Set enviroment vars to run tests"""
        self.monkeypatch.setattr("mezcla.tpo_common.env_options", {
            'VAR_STRING': 'this is a string variable',
            'ANOTHER_VAR': 'this is another env. var.'
        })
        self.monkeypatch.setattr("mezcla.tpo_common.env_defaults", {
            'VAR_STRING': 'empty',
            'ANOTHER_VAR': '2022'
        })

        
    @pytest.mark.xfail                   # TODO: remove xfail
    @trap_exception
    def test_formatted_environment_option_descriptions(self):
        """Ensure formatted_environment_option_descriptions works as expected"""
        debug.trace(4, "test_formatted_environment_option_descriptions()")

        self.set_test_env_var()

        # Test sort
        expected = (
            'VAR_STRING\tthis is a string variable (empty)\n'
            '\tANOTHER_VAR\tthis is another env. var. (2022)'
        )
        self.do_assert(THE_MODULE.formatted_environment_option_descriptions(sort=False) == expected)
        expected = (
            'ANOTHER_VAR\tthis is another env. var. (2022)\n'
            '\tVAR_STRING\tthis is a string variable (empty)'
        )
        self.do_assert(THE_MODULE.formatted_environment_option_descriptions(sort=True) == expected)

        # Test include_all
        # NOTE: this is being tested on test_system.test_get_environment_option_descriptions()

        # Test indent
        expected = (
            'VAR_STRING + this is a string variable (empty)\n'
            ' + ANOTHER_VAR + this is another env. var. (2022)'
        )
        self.do_assert(THE_MODULE.formatted_environment_option_descriptions(indent=' + ') == expected)

    def test_get_registered_env_options(self):
        """Ensure get_registered_env_options works as expected"""
        debug.trace(4, "test_get_registered_env_options()")
        self.set_test_env_var()
        result = THE_MODULE.get_registered_env_options()
        assert isinstance(result, list)
        assert 'VAR_STRING' in result
        assert len(result) == 2

    @trap_exception
    def test_get_environment_option_descriptions(self):
        """Test get_environment_option_descriptions"""
        debug.trace(4, "test_get_environment_option_descriptions()")
        self.set_test_env_var()
        result = THE_MODULE.get_environment_option_descriptions(include_default=True)
        self.do_assert(isinstance(result, list))
        self.do_assert("(2022)" in str(result), "default added")
        self.do_assert(len(result) == 2)

    @pytest.mark.xfail
    def test_getenv_real(self):
        """Ensure getenv_real works as expected"""
        debug.trace(4, "test_getenv_real()")
        self.monkeypatch.setenv('TEST_NEG_VAR', "-123_456")
        neg_env = THE_MODULE.getenv_real('TEST_NEG_VAR', default=2.71)

        self.monkeypatch.setenv('TEST_POS_VAR', "123.45")
        pos_env = THE_MODULE.getenv_real('TEST_POS_VAR', default=2.71)

        assert neg_env == -123456
        assert pos_env == 123.45

    @pytest.mark.xfail
    def test_getenv_float(self):
        """Ensure getenv_float works as expected"""
        debug.trace(4, "test_getenv_float()")
        self.monkeypatch.setenv('TEST_INF_VAR', "inf")
        inf_env = THE_MODULE.getenv_float('TEST_INF_VAR', default=2.71)

        self.monkeypatch.setenv('TEST_NAN_VAR', 'nan')
        nan_env = THE_MODULE.getenv_float('TEST_NAN_VAR', default=2.71)

        self.monkeypatch.setenv('TEST_NEG_VAR', 'a123')
        default_env = THE_MODULE.getenv_float('TEST_NEG_VAR', default=2.71)

        assert inf_env == math.inf
        assert math.isnan(nan_env)
        assert default_env == 2.71

    def test_get_current_function_name(self):
        """Test(s) for get_current_function_name()"""
        assert THE_MODULE.get_current_function_name() == "test_get_current_function_name"
        return

    @pytest.mark.xfail
    def test_get_property_value(self):
        """Ensure get_property_value works as expected"""
        debug.trace(4, "test_get_property_value()")
        class test_object:
            """ test class """
            def __init__(self):
                self.property = "test"

        test_obj = test_object()

        assert THE_MODULE.get_property_value(test_obj, 'property') == "test"
        assert THE_MODULE.get_property_value(test_obj, 'non_prop', None) is None

    @pytest.mark.xfail
    def test_simple_format(self):
        """Ensure simple_format works as expected"""
        debug.trace(4, "test_simple_format()")
        fubar = 202

        assert THE_MODULE.simple_format("{F} vs. {f}", namespace={"F": FUBAR, "f": fubar}) == f"{FUBAR} vs. {fubar}"
        assert THE_MODULE.simple_format("{{F}} vs. {{f}}", namespace={"F":FUBAR, "f":fubar}) == r"{F} vs. {f}"

        THE_MODULE.simple_format("{F} vs. {f}", namespace={})
        stderr = self.get_stderr()
        assert r"Unable to resolve replacement 'F' in '{F} vs. {f}'" in stderr

    def test_format(self):
        """Ensure format resolves from local and global namespace, and that local takes precedence"""
        fubar = 202
        assert THE_MODULE.format("{F} vs. {f}", F=FUBAR, f=fubar) == ("%s vs. %s" % (FUBAR, fubar))
        # pylint: disable=redefined-outer-name
        FOOBAR = 21
        assert THE_MODULE.format("{FOO}", FOO=FOOBAR) == str(FOOBAR)
        ## TODO: assert "Hey Jos\xc3\xa9" == THE_MODULE.format("Hey {j}", j=JOSE)
        return

    @pytest.mark.xfail
    def test_init_logging(self):
        """Ensure init_logging works as expected"""
        debug.trace(4, "test_init_logging()")
        THE_MODULE.init_logging()
        stderr = self.get_stderr()
        assert "init_logging" in stderr

    def test_load_object(self):
        """Ensure load_object works as expected"""
        debug.trace(4, "test_load_object()")
        test_dict = {
            1: 'first',
            2: 'second',
        }
        test_filename = self.get_temp_file()
        test_file = open(test_filename, 'wb')
        pickle.dump(test_dict, test_file)
        test_file.close()

        assert THE_MODULE.load_object(test_filename) == test_dict

    def test_store_object(self):
        """Ensure store_object works as expected"""
        debug.trace(4, "test_store_object()")
        test_dict = {
            1: 'first',
            2: 'second',
        }
        test_filename = self.get_temp_file()

        THE_MODULE.store_object(test_filename, test_dict)

        test_file = open(test_filename, 'rb')
        actual_object = pickle.load(test_file)
        assert actual_object == test_dict
        test_file.close()

    def test_dump_stored_object(self):
        """Ensure dump_stored_object works as expected"""
        debug.trace(4, "test_dump_stored_object()")

        ## TODO: add tests related to redirect_stderr and restore_stderr

        test_dict = {
            1: 'first',
            2: 'second',
        }
        test_filename = self.get_temp_file()

        THE_MODULE.store_object(test_filename, test_dict)

        test_file = open(test_filename, 'rb')
        actual_object = pickle.load(test_file)
        assert actual_object == test_dict
        test_file.close()

    @pytest.mark.xfail
    def test_create_lookup_table(self):
        """Ensure create_lookup_table works as expected"""
        debug.trace(4, "test_create_lookup_table()")
        content = (
            "a\t1st\n"
            "b\t2nd\n"
            "c\n"
        )

        temp_file = self.get_temp_file()
        system.write_file(temp_file, content)

        linum_table = THE_MODULE.create_lookup_table(temp_file, use_linenum=True)
        table = THE_MODULE.create_lookup_table(temp_file, use_linenum=False)

        assert table['a'] == '1st'
        assert table['b'] == '2nd'
        assert linum_table[3] == 'c'

    def test_lookup_key(self):
        """Ensure lookup_key works as expected"""
        debug.trace(4, "test_lookup_key()")
        test_table = {
            'first': '1st',
            'second': '2nd',
        }
        assert THE_MODULE.lookup_key(test_table, 'second', 'two-nd') == '2nd'
        assert THE_MODULE.lookup_key(test_table, 'third', '3rd') == '3rd'

    def test_create_boolean_lookup_table(self):
        """Ensure create_boolean_lookup_table works as expected"""
        debug.trace(4, "test_create_boolean_lookup_table()")

        content = (
            'EmailEntered - someemail@hotmail.com\n'
            'PasswdEntered - 12345\n'
            'IsBusiness - True\n'
        )
        expected = {
            'emailentered - someemail@hotmail.com': True,
            'passwdentered - 12345': True,
            'isbusiness - true': True,
        }

        temp_file = self.get_temp_file()
        system.write_file(temp_file, content)
        assert THE_MODULE.create_boolean_lookup_table(temp_file) == expected

    @pytest.mark.xfail
    def test_normalize_frequencies(self):
        """Ensure normalize_frequencies works as expected"""
        debug.trace(4, "test_normalize_frequencies()")
        dictionary = {'a':25, 'b': 55, 'c': 20}
        THE_MODULE.normalize_frequencies(dictionary)
        assert dictionary == {'a': 0.25, 'b': 0.55, 'c': 0.2}

    @pytest.mark.xfail
    def test_sort_frequencies(self):
        """Ensure sort_frequencies works as expected"""
        debug.trace(4, "test_sort_frequencies()")
        dictionary = {'a': 0.25, 'b': 0.55, 'c': 0.2}
        assert THE_MODULE.sort_frequencies(dictionary) == ['b', 'a', 'c']

    def test_sort_weighted_hash(self):
        """Ensure sort_weighted_hash works as expected"""
        debug.trace(4, "test_sort_weighted_hash()")
        test_hash = {
            'bananas': 3,
            'apples': 1411,
            'peach': 43,
        }
        sorted_hash = [
            ('bananas', 3),
            ('peach', 43),
            ('apples', 1411),
        ]
        reversed_hash = [
            ('apples', 1411),
            ('peach', 43),
            ('bananas', 3),
        ]
        assert THE_MODULE.sort_weighted_hash(test_hash) == reversed_hash
        assert THE_MODULE.sort_weighted_hash(test_hash, reverse=False) == sorted_hash
        assert len(THE_MODULE.sort_weighted_hash(test_hash, max_num=2)) == 2

    @pytest.mark.xfail
    def test_format_freq_hash(self):
        """Ensure format_freq_hash works as expected"""
        debug.trace(4, "test_format_freq_hash()")
        dictionary = {'a': 2.25, 'b': 3.55, 'c': 1.2}
        expected_output = (
            'test_format_freq_hash\n'
            '\tb\t4\n'
            '\ta\t2\n'
            '\tc\t1'
        )
        formatted_dict = THE_MODULE.format_freq_hash(dictionary, 'test_format_freq_hash')
        assert formatted_dict == expected_output

    def test_union(self):
        """Ensure union works as expected"""
        debug.trace(4, "test_union()")
        assert THE_MODULE.union([1, 2, 3], [2, 3, 4, 5]) == [1, 2, 3, 4, 5]

    def test_intersection(self):
        """Ensure intersection works as expected"""
        debug.trace(4, "test_intersection()")
        assert THE_MODULE.intersection([1, 2, 3, 4, 5], [2, 4]) == [2, 4]

    def test_is_subset(self):
        """Ensure is_subset works as expected"""
        debug.trace(4, "test_is_subset()")
        assert THE_MODULE.is_subset(['mouse', 'dog'], ['dog', 'cat', 'mouse'])

    def test_difference(self):
        """Ensures set difference works as expected"""
        assert THE_MODULE.difference([5, 4, 3, 2, 1], [4, 2]) == [1, 3, 5]
        assert THE_MODULE.difference([1, 2, 3], [2]) == [1, 3]
        assert THE_MODULE.difference([1, 1, 2, 2], [1]) == [2]

    def test_remove_all(self):
        """Ensure remove_all works as expected"""
        debug.trace(4, "test_remove_all()")
        assert THE_MODULE.remove_all([5, 4, 3, 2, 1], [4, 2, 0]) == [5, 3, 1]
        assert THE_MODULE.remove_all(['A', 'B', 'C', 'D'], ['A', 'B', 'D']) == ['C']
        assert THE_MODULE.remove_all(['a', 'B', 'c', 'D'], ['A', 'b', 'd'], ignore_case=True) == ['c']

    def test_equivalent(self):
        """Ensure equivalent works as expected"""
        debug.trace(4, "test_equivalent()")
        assert THE_MODULE.equivalent([1, 2, 3], [1, 2, 3])
        assert not THE_MODULE.equivalent([1, 2, 3, 4], [1, 2, 3])
        assert not THE_MODULE.equivalent([1, 3], [1, 2])

    def test_append_new(self):
        """Ensure append_new works as expected"""
        debug.trace(4, "test_append_new()")
        assert THE_MODULE.append_new([1, 2], 3) == [1, 2, 3]
        assert THE_MODULE.append_new([1, 2, 3], 3) == [1, 2, 3]

    def test_extract_list(self):
        """Ensure extract_list works as expected"""
        debug.trace(4, "test_extract_list()")
        assert THE_MODULE.extract_list('a,b,c') == ['a', 'b', 'c']

    def test_is_subsumed(self):
        """Ensure is_subsumed works as expected"""
        debug.trace(4, "test_is_subsumed()")
        assert THE_MODULE.is_subsumed("dog", ["dog house", "catnip"])
        assert not THE_MODULE.is_subsumed("cat", ["dog house", "catnip"])

    def test_round_num(self):
        """Ensure round_num works as expected"""
        debug.trace(4, "test_round_num()")
        assert THE_MODULE.round_num(15000, 3) == "15000.000"
        assert THE_MODULE.round_num(15000, 3, False) == "15000"

    def test_round_nums(self):
        """Ensure round_nums works as expected"""
        debug.trace(4, "test_round_nums()")
        assert THE_MODULE.round_nums([0.333333, 0.666666, 0.99999]) == ['0.333', '0.667', '1.000']

    def test_round(self):
        """Ensure round works as expected"""
        debug.trace(4, "test_round()")
        assert THE_MODULE.round(15000) == 15000.0
        assert THE_MODULE.round([0.333333, 0.666666, 0.99999]) == [0.333, 0.667, 1.0]

    def test_normalize(self):
        """Ensure normalize works as expected"""
        debug.trace(4, "test_normalize()")
        assert THE_MODULE.normalize([1, 2, 3]) == [0.0, 0.5, 1.0]

    def test_is_numeric(self):
        """Ensure is_numeric works as expected"""
        debug.trace(4, "test_is_numeric()")
        assert THE_MODULE.is_numeric("123")
        assert not THE_MODULE.is_numeric("one")

    def test_safe_int(self):
        """Ensure safe_int works as expected"""
        debug.trace(4, "test_safe_int()")
        assert THE_MODULE.safe_int('1') == 1
        assert THE_MODULE.safe_int(2.0) == 2
        assert THE_MODULE.safe_int("F", base=16) == 15
        assert THE_MODULE.safe_int("77", base=8) == 63
        assert THE_MODULE.safe_int("four") == 0

    def test_safe_float(self):
        """Ensure safe_float works as expected"""
        debug.trace(4, "test_safe_float()")
        assert THE_MODULE.safe_float(5) == 5.0
        assert THE_MODULE.safe_float("3") == 3.0
        assert THE_MODULE.safe_float("three") == 0.0

    @pytest.mark.xfail
    def test_reference_variables(self):
        """Ensure reference_variables works as expected"""
        debug.trace(4, "test_reference_variables()")
        self.monkeypatch.setattr("mezcla.debug.trace_level", 10)
        a = "dummy"
        ## BAD: THE_MODULE.reference_variables("\'a\'")
        THE_MODULE.reference_variables(a)

        stderr = self.get_stderr()
        ## BAD: assert 'reference_variables("\'a\'",)' in stderr
        # note: tpo_common calls into debug version 
        assert f"reference_var('{a}',)" in stderr

    @pytest.mark.xfail
    def test_memodict(self):
        """Ensure memodict works as expected"""
        debug.trace(4, "test_memodict()")
        self.patch_trace_level(4)

        @THE_MODULE.memodict
        def add_1(x:int)-> int:
            debug.trace(4, f"add_1({x})")
            return x + 1

        _ = add_1(1)
        stderr = self.get_stderr()
        assert 'add_1(1)' in stderr

        self.clear_stderr()
        _ = add_1(1)
        stderr = self.get_stderr()
        assert 'add_1(1)' not in stderr

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_exit(self):
        """Ensure exit works as expected"""
        debug.trace(4, "test_exit()")
        def sys_exit_mock():
            return 'exit'
        self.monkeypatch.setattr(sys, "exit", sys_exit_mock)
        assert THE_MODULE.exit('test exit method') == 'exit'
        # Exit is mocked, ignore code editor hiding
        ## TODO: for some reason (probably the debug level) the message is not being printed
        captured = self.get_stdout_stderr()
        debug.trace_object(5, captured)
        ## TODO: assert "test exit method" in captured.err
        debug.assertion("test exit method" in captured[1])

    def test_dummy_main(self):
        """Ensure dummy_main works as expected"""
        debug.trace(4, "test_dummy_main()")
        THE_MODULE.dummy_main()
        captured = self.get_stdout()
        assert 'Environment options' in captured

    @trap_exception
    def test_getenv(self):
        """Ensure getenv works as expected"""
        debug.trace(4, "test_getenv()")
        self.monkeypatch.setenv('TEST_ENV_VAR', 'some value', prepend=False)
        assert THE_MODULE.getenv('TEST_ENV_VAR') == 'some value'

    def test_getenv_value(self):
        """Ensure getenv_value works as expected"""
        debug.trace(4, "test_getenv_value()")
        self.monkeypatch.setenv('NEW_ENV_VAR', 'some value', prepend=False)
        assert THE_MODULE.getenv_value('NEW_ENV_VAR', default='empty', description='another test env var') == 'some value'
        assert THE_MODULE.env_defaults['NEW_ENV_VAR'] == 'empty'
        assert THE_MODULE.env_options['NEW_ENV_VAR'] == 'another test env var'

    def test_getenv_text(self):
        """Ensure getenv_text works as expected"""
        debug.trace(4, "test_getenv_text()")
        self.monkeypatch.setenv('TEST_ENV_VAR', 'some value', prepend=False)
        assert THE_MODULE.getenv_text('TEST_ENV_VAR') == 'some value'
        assert THE_MODULE.getenv_text("REALLY FUBAR?", False) == 'False'

    def test_getenv_number(self):
        """Ensure getenv_number works as expected"""
        debug.trace(4, "test_getenv_number()")
        self.monkeypatch.setenv('TEST_NUMBER', '9.81', prepend=False)
        assert THE_MODULE.getenv_number('TEST_NUMBER', default=20) == 9.81
        assert THE_MODULE.getenv_number("REALLY FUBAR", 123) == 123.0

    def test_getenv_int(self):
        """Ensure getenv_int works as expected"""
        debug.trace(4, "test_getenv_int()")
        self.monkeypatch.setenv('TEST_NUMBER', '34', prepend=False)
        assert THE_MODULE.getenv_int('TEST_NUMBER', default=20) == 34
        assert THE_MODULE.getenv_int("REALLY FUBAR", 123) == 123

    def test_getenv_bool(self):
        """Ensure getenv_bool works as expected"""
        debug.trace(4, "test_getenv_bool()")
        self.monkeypatch.setenv('TEST_BOOL', 'FALSE', prepend=False)
        assert not THE_MODULE.getenv_bool('TEST_BOOL', None)
        self.monkeypatch.setenv('TEST_BOOL', '  true   ', prepend=False)
        assert THE_MODULE.getenv_bool('TEST_BOOL', None)
        assert not isinstance(THE_MODULE.getenv_boolean("REALLY FUBAR?", None), bool)
        assert isinstance(THE_MODULE.getenv_boolean("REALLY FUBAR?", False), bool)


## BAD (needs to use monkeypatch):
## def set_test_env_var():
##     """Set enviroment vars to run tests"""
##     THE_MODULE.env_options = {
##         'VAR_STRING': 'this is a string variable',
##         'ANOTHER_VAR': 'this is another env. var.'
##     }
##     THE_MODULE.env_defaults = {
##         'VAR_STRING': 'empty',
##         'ANOTHER_VAR': '2022'
##     }


if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)
